
open Base
open Ast

exception TypeError of string * location

module Key = struct type t = string let compare = compare end
module Env = Map.Make(Key)

type profil = asa_typ list * asa_typ
type infoid = Typ of asa_typ | Prof of profil

let add_var x t vis = Env.add x (Typ t) vis
    
let add_fun f p vis = Env.add f (Prof p) vis

let rec type_expr vis e =
  match e.asa with
  | Cst (Int _) -> TypInt
  | Cst (Bool _) -> TypBool
  | Var x -> 
      if Env.mem x vis then 
        match Env.find x vis with 
          | (Typ t) -> t 
          | (Prof _) -> raise (TypeError ("function used as ident",e.loc))
      else raise (TypeError ("ident not declared",e.loc))
  | Binop(op,e1,e2) when List.mem op [Plus;Mult;Div;Minus] ->
      let tau1 = type_expr vis e1 in
      let tau2 = type_expr vis e2 in
        if tau1 <> TypInt then 
	  raise (TypeError ("Expecting expression of type int",e1.loc))
        else if tau2 <> TypInt then 
	  raise (TypeError ("Expecting expression of type int",e2.loc))
	else TypInt
  | Binop(Eq,e1,e2) ->  
      let tau1 = type_expr vis e1 
      and tau2 = type_expr vis e2 in
        if tau1 <> tau2 then 
	  raise (TypeError ("Expecting expression of equal types",e.loc))
	else TypBool
  | Binop(rel,e1,e2) ->
      let tau1 = type_expr vis e1 
      and tau2 = type_expr vis e2 in
        if tau1 <> TypInt then 
          raise (TypeError ("Expecting expression of type int",e1.loc))
        else if tau2 <> TypInt then 
          raise (TypeError ("Expecting expression of type int",e2.loc))
        else TypBool
  | Letin(x,e1,e2)  ->
      let tau1 = type_expr vis e1 in
      let vis' = add_var x tau1 vis in
      let tau2 = type_expr vis' e2 in
        tau2
  | If(e1,e2,e3) ->
      let tau1 = type_expr vis e1 in
      if tau1<>TypBool then
        raise (TypeError ("Expecting expression of type int",e1.loc))
      else
        let tau2 = type_expr vis e2 in
        let tau3 = type_expr vis e3 in
          if tau2 <> tau3 then 
            raise (TypeError ("Expecting expression of equal types",e.loc))
          else tau2
  | Call(f,le) ->   
      if Env.mem f vis then 
        match Env.find f vis with 
          | Prof(lt,t) -> 
              type_args e.loc vis (le,lt) ;
              t
          | Typ _ -> raise (TypeError ("function expected",e.loc))
      else raise (TypeError ("ident not declared",e.loc))

and type_args loc vis = function 
  | [],[] -> ()
  | e::le,t::lt ->
      let t' = type_expr vis e in
      if t<>t' then raise (TypeError ("wrong type for argument",e.loc))
      else type_args loc vis (le,lt)
  | _ -> raise (TypeError ("wrong number of arguments",loc))

let rec check_decls vis = function 
  | [] -> ()
  | {asa=Vdecl(id,e)} :: p -> 
      let tau = type_expr vis e in 
      let vis' = add_var id tau vis in 
        check_decls vis' p 
  | {asa=Fdecl(id,lv,tr,e)} :: p -> 
      let prof = (List.map snd lv,tr) in
      let vis' = add_fun id prof vis in
      let vispar = 
        List.fold_right 
          (fun (x,t) v -> add_var x t v)
          lv vis'
      in 
      let tau' = type_expr vispar e in
        if tau'<>tr then raise (TypeError ("wrong result type",e.loc))
        else check_decls vis' p 

let check_program p = check_decls Env.empty p
